from os import walk
from os.path import join, isfile, isdir


from click.testing import CliRunner
import pandas as pd
import numpy as np
from openscm.scmdataframe import ScmDataFrame

import netcdf_scm
from netcdf_scm.cli import crunch_data


from conftest import TEST_DATA_KNMI_DIR, TEST_DATA_MARBLE_CMIP5_DIR


def test_crunching(tmpdir):
    INPUT_DIR = TEST_DATA_MARBLE_CMIP5_DIR
    OUTPUT_DIR = str(tmpdir)
    VAR_TO_CRUNCH = ".*tas.*"

    runner = CliRunner()
    result = runner.invoke(
        crunch_data,
        [
            INPUT_DIR,
            OUTPUT_DIR,
            "--cube-type",
            "MarbleCMIP5",
            "--var-to-crunch",
            VAR_TO_CRUNCH,
            "-f",
        ],
    )
    assert result.exit_code == 0
    assert "NetCDF SCM version: {}".format(netcdf_scm.__version__) in result.output
    assert "Making output directory:" in result.output

    THRESHOLD_PERCENTAGE_DIFF = 10 ** -1
    files_found = 0
    for dirpath, dirnames, filenames in walk(OUTPUT_DIR):
        if not dirnames:
            assert len(filenames) == 1
            filename = filenames[0]
            files_found += 1

            knmi_data_name = "global_{}.dat".format("_".join(filename.split("_")[1:6]))
            knmi_data_path = join(TEST_DATA_KNMI_DIR, knmi_data_name)

            if not isfile(knmi_data_path):
                print("No data available for {}".format(knmi_data_path))
                continue

            knmi_data = pd.read_csv(
                knmi_data_path,
                skiprows=3,
                delim_whitespace=True,
                header=None,
                names=["year", 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
            ).melt(id_vars="year", var_name="month")
            knmi_data["year"] = knmi_data["year"].astype(int)
            knmi_data["month"] = knmi_data["month"].astype(int)
            knmi_data = knmi_data.set_index(["year", "month"])

            crunched_data = ScmDataFrame(join(dirpath, filename))
            comparison_data = (
                crunched_data.filter(region="World")
                .timeseries()
                .stack()
                .to_frame()
                .reset_index()[["time", 0]]
            )
            comparison_data = comparison_data.rename({0: "value"}, axis="columns")

            comparison_data["year"] = comparison_data["time"].apply(lambda x: x.year)
            comparison_data["month"] = comparison_data["time"].apply(lambda x: x.month)

            comparison_data = comparison_data.drop("time", axis="columns")
            comparison_data = comparison_data.set_index(["year", "month"])

            rel_difference = (knmi_data - comparison_data) / knmi_data
            # drop regions where times are not equal
            rel_difference = rel_difference.dropna()
            assert not rel_difference.empty, "not testing anything"

            assert_message = "{} data is not the same to within {}%".format(
                filename, THRESHOLD_PERCENTAGE_DIFF
            )
            all_close = (
                np.abs(rel_difference.values) < THRESHOLD_PERCENTAGE_DIFF / 100
            ).all()
            assert all_close, assert_message

            print(
                "{} file matches KNMI data to within {}%".format(
                    filename, THRESHOLD_PERCENTAGE_DIFF
                )
            )

    assert files_found == 5


def test_crunching_arguments(tmpdir):
    INPUT_DIR = TEST_DATA_MARBLE_CMIP5_DIR
    OUTPUT_DIR = str(tmpdir)
    VAR_TO_CRUNCH = ".*fco2antt.*"
    DATA_SUB_DIR = "custom-name"
    LAND_MASK_TRESHHOLD = 45

    runner = CliRunner()
    result = runner.invoke(
        crunch_data,
        [
            INPUT_DIR,
            OUTPUT_DIR,
            "--cube-type",
            "MarbleCMIP5",
            "--var-to-crunch",
            VAR_TO_CRUNCH,
            "--data-sub-dir",
            DATA_SUB_DIR,
            "--land-mask-threshold",
            LAND_MASK_TRESHHOLD,
            "-f",
        ],
    )
    assert result.exit_code == 0

    assert "NetCDF SCM version: {}".format(netcdf_scm.__version__) in result.output
    assert "Making output directory:" in result.output

    assert "fco2antt" in result.output
    assert "tas" not in result.output

    assert "Making output directory:" in result.output
    assert isdir(join(OUTPUT_DIR, DATA_SUB_DIR, "cmip5"))

    assert "land-mask-threshold: {}".format(LAND_MASK_TRESHHOLD) in result.output

    result_skip = runner.invoke(
        crunch_data,
        [
            INPUT_DIR,
            OUTPUT_DIR,
            "--cube-type",
            "MarbleCMIP5",
            "--var-to-crunch",
            VAR_TO_CRUNCH,
            "--data-sub-dir",
            DATA_SUB_DIR,
            "--land-mask-threshold",
            LAND_MASK_TRESHHOLD,
        ],
    )
    assert result_skip.exit_code == 0

    skip_str = (
        "Skipped (already exist, not overwriting)\n"
        "========================================\n"
        "- {}".format(
            join(
                OUTPUT_DIR,
                DATA_SUB_DIR,
                "cmip5",
                "1pctCO2",
                "Amon",
                "fco2antt",
                "CanESM2",
                "r1i1p1",
                "netcdf-scm_fco2antt_Amon_CanESM2_1pctCO2_r1i1p1_185001-198912.csv",
            )
        )
    )
    assert skip_str in result_skip.output


def test_crunching_other_cube(tmpdir):
    INPUT_DIR = TEST_DATA_MARBLE_CMIP5_DIR
    OUTPUT_DIR = str(tmpdir)
    CUBE = "CMIP6Output"

    runner = CliRunner()
    result = runner.invoke(crunch_data, [INPUT_DIR, OUTPUT_DIR, "--cube-type", CUBE])
    assert result.exit_code  # non-zero exit code

    assert "cube-type: {}".format(CUBE) in result.output



"""Command line interface
"""
from os import walk, makedirs, path
from os.path import join, isfile
import re
import warnings
import traceback
import time
from time import gmtime, strftime

import click
import netcdf_scm
from netcdf_scm.iris_cube_wrappers import (
    SCMCube,
    MarbleCMIP5Cube,
    CMIP6Input4MIPsCube,
    CMIP6OutputCube,
)
import progressbar


_CUBES = {
    "Scm": SCMCube,
    "MarbleCMIP5": MarbleCMIP5Cube,
    "CMIP6Input4MIPs": CMIP6Input4MIPsCube,
    "CMIP6Output": CMIP6OutputCube,
}


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.argument("src", type=click.Path(exists=True, readable=True, resolve_path=True))
@click.argument(
    "dst", type=click.Path(file_okay=False, writable=True, resolve_path=True)
)
@click.option(
    "--cube-type",
    default="Scm",
    type=click.Choice(["Scm", "MarbleCMIP5", "CMIP6Input4MIPs", "CMIP6Output"]),
    show_default=True,
    help="Cube to use for crunching.",
)
@click.option(
    "--var-to-crunch",
    default=".*",
    show_default=True,
    help="Variable to crunch (uses regexp syntax, matches on filepaths).",
)
@click.option(
    "--land-mask-threshold",
    default=50.0,
    show_default=True,
    help="Minimum land fraction for a box to be considered land.",
)
@click.option(
    "--data-sub-dir",
    default="netcdf-scm-crunched",
    show_default=True,
    help="Sub-directory of ``dst`` to save data in.",
)
@click.option(
    "--force/--do-not-force",
    "-f",
    help="Overwrite any existing files.",
    default=False,
    show_default=True,
)
def crunch_data(
    src, dst, cube_type, var_to_crunch, land_mask_threshold, data_sub_dir, force
):
    """Crunch data in ``src`` to OpenSCM csv's in ``dst``.

    ``src`` is searched recursively and netcdf-scm will attemp to crunch all the files
    found. The directory structure in ``src`` will be mirrored in ``dst``.

    Failures and warnings are recorded and written into a text file in ``dst``.
    """
    title = "NetCDF Crunching"
    output_prefix = "netcdf-scm"
    separator = "_"
    timestamp = strftime("%Y%m%d %H%M%S", gmtime())
    out_dir = join(dst, data_sub_dir)

    metadata_header = (
        "{}\n"
        "{}\n"
        "NetCDF SCM version: {}\n"
        "\n"
        "time: {}\n"
        "cube-type: {}\n"
        "source: {}\n"
        "destination: {}\n"
        "var-to-crunch: {}\n"
        "land-mask-threshold: {}\n"
        "force: {}\n\n"
        "".format(
            title,
            "=" * len(title),
            netcdf_scm.__version__,
            timestamp,
            cube_type,
            src,
            out_dir,
            var_to_crunch,
            land_mask_threshold,
            force,
        )
    )
    click.echo(metadata_header)

    if not path.exists(out_dir):
        click.echo("Making output directory: {}\n".format(out_dir))
        makedirs(out_dir)

    already_exist_files = []

    var_regexp = re.compile(var_to_crunch)

    time.sleep(0.5)  # needed to get logging bar in right place...
    # really should use a logger here
    with warnings.catch_warnings(record=True) as recorded_warns:
        failures = []
        format_custom_text = progressbar.FormatCustomText(
            "Current directory :: %(curr_dir)-400s", {"curr_dir": "uninitialised"}
        )
        bar = progressbar.ProgressBar(
            widgets=[progressbar.SimpleProgress(), ". ", format_custom_text],
            max_value=len([w for w in walk(src)]),
            prefix="Visiting directory ",
        ).start()
        for i, (dirpath, dirnames, filenames) in enumerate(walk(src)):
            if not dirnames:
                if not var_regexp.match(dirpath):
                    continue
                format_custom_text.update_mapping(curr_dir=dirpath)
                bar.update(i)
                scmcube = _CUBES[cube_type]()
                try:
                    if len(filenames) == 1:
                        out_filename = separator.join(
                            [output_prefix, filenames[0].replace(".nc", ".csv")]
                        )
                        scmcube.load_data_from_path(join(dirpath, filenames[0]))
                        out_filedir = scmcube._get_data_directory().replace(
                            scmcube.root_dir, out_dir
                        )
                        out_filepath = join(out_filedir, out_filename)
                        if not path.exists(out_filedir):
                            makedirs(out_filedir)
                        if not force and isfile(out_filepath):
                            already_exist_files.append(out_filepath)
                            continue
                    else:
                        scmcube.load_data_in_directory(dirpath)
                        out_filename = separator.join(
                            [
                                output_prefix,
                                scmcube._get_data_filename().replace(".nc", ".csv"),
                            ]
                        )

                        out_filedir = scmcube._get_data_directory().replace(
                            scmcube.root_dir, out_dir
                        )
                        out_filepath = join(out_filedir, out_filename)
                        if not path.exists(out_filedir):
                            makedirs(out_filedir)

                        if not force and isfile(out_filepath):
                            already_exist_files.append(out_filepath)
                            continue

                    results = scmcube.get_scm_timeseries(
                        land_mask_threshold=land_mask_threshold
                    )
                    results.to_csv(out_filepath)

                except Exception:
                    header = "Exception"
                    exc_string = (
                        header
                        + "\n"
                        + "-" * len(header)
                        + "\n"
                        + traceback.format_exc()
                    )

                    # ideally would write to a logger here
                    failures.append("{}\n{}\n{}".format(dirpath, filenames, exc_string))
                    continue

        bar.finish()

    header_underline = "="
    msg_underline = "-"

    warnings_together = "\n\n{}\n\n".format(15 * msg_underline).join(
        [str(rw.message) for rw in recorded_warns]
    )

    warnings_header = "Warnings"
    warnings_string = "{}\n{}\n{}".format(
        warnings_header, len(warnings_header) * header_underline, warnings_together
    )

    failures_header = "Failures"
    failures_string = "{}\n{}\n{}".format(
        failures_header,
        len(failures_header) * header_underline,
        msg_underline.join(failures),
    )

    already_exist_header = "Skipped (already exist, not overwriting)"
    if already_exist_files:
        already_exist_files_string = "- {}\n".format("\n- ".join(already_exist_files))
    else:
        already_exist_files_string = ""
    already_exist_string = "{}\n{}\n{}".format(
        already_exist_header,
        len(already_exist_header) * header_underline,
        already_exist_files_string,
    )

    output_string = "\n\n{}\n\n{}\n\n{}\n\n{}".format(
        metadata_header, failures_string, warnings_string, already_exist_string
    )
    click.echo(output_string)
    summary_file = join(
        out_dir,
        "{}-failures-and-warnings.txt".format(
            timestamp.replace(" ", "_").replace(":", "")
        ),
    )
    with open(summary_file, "w") as ef:
        ef.write(output_string)

    if failures:
        raise click.ClickException("Failures were found")



"""This module contains our wrappers of the iris cube.

These classes automate handling of a number of netCDF processing steps.
For example, finding surface land fraction files, applying masks to data and
returning timeseries in key regions for simple climate models.
"""
import os
from os.path import join, dirname, basename, splitext
import re
import warnings
import traceback
from datetime import datetime
from dateutil.relativedelta import relativedelta
from dateutil import parser


import numpy as np
import pandas as pd
from openscm.scmdataframe import ScmDataFrame

try:
    import iris
    from iris.util import broadcast_to_shape, unify_time_units
    import iris.analysis.cartography
    import iris.experimental.equalise_cubes
    from iris.exceptions import CoordinateNotFoundError
    import cftime
    import cf_units
except ModuleNotFoundError:
    from .errors import raise_no_iris_warning

    raise_no_iris_warning()


from .utils import (
    get_cube_timeseries_data,
    get_scm_cube_time_axis_in_calendar,
    assert_all_time_axes_same,
    take_lat_lon_mean,
    apply_mask,
    unify_lat_lon,
)


class SCMCube(object):
    """Class for processing netCDF files for use in simple climate models.

    Common, shared operations are implemented here.
    However, methods like ``_get_data_directory`` raise ``NotImplementedError``
    because these are always context dependent.
    Hence to use this base class, you must use a subclass of it which defines these
    context specific methods.
    """

    sftlf_var = "sftlf"
    """str: The name of the variable associated with the land-surface fraction in each gridbox.

    If required, this is used when looking for the land-surface fraction file which
    belongs to a given data file. For example, if our data file is
    ``tas_Amon_HadCM3_rcp45_r1i1p1_200601.nc`` then ``sftlf_var`` can be used to work
    out the name of the associated land-surface fraction file. In some cases, it might
    be as simple as replacing ``tas`` with the value of ``sftlf_var``.
    """

    areacella_var = "areacella"
    """str: The name of the variable associated with the area of each gridbox.

    If required, this is used to determine the area of each cell in a data file. For
    example, if our data file is ``tas_Amon_HadCM3_rcp45_r1i1p1_200601.nc`` then
    ``areacella_var`` can be used to work  out the name of the associated cell area
    file. In some cases, it might be as simple as replacing ``tas`` with the value of
    ``areacella_var``.
    """

    time_name = "time"
    """str: The expected name of the time co-ordinate in data."""

    lat_name = "latitude"
    """str: The expected name of the latitude co-ordinate in data."""

    lon_name = "longitude"
    """str: The expected name of the longitude co-ordinate in data."""

    time_period_separator = "-"
    """str: Character used to separate time period strings in the time period indicator in filenames.

    e.g. ``-`` is the 'time period separator' in "2015-2030".
    """

    _time_period_regex = None

    _known_timestamps = {
        4: {"datetime_str": "%Y", "expected_timestep": relativedelta(years=+1)},
        6: {"datetime_str": "%Y%m", "expected_timestep": relativedelta(months=+1)},
        8: {"datetime_str": "%Y%m%d", "expected_timestep": relativedelta(days=+1)},
        10: {"datetime_str": "%Y%m%d%H", "expected_timestep": relativedelta(hours=+1)},
    }
    _timestamp_definitions = None

    @property
    def time_period_regex(self):
        """:obj:`_sre.SRE_Pattern`: Regular expression which captures the timeseries identifier in input data files.

        For help on regular expressions, see :ref:`regular expressions <regular-expressions>`.
        """

        if self._time_period_regex is None:
            self._time_period_regex = re.compile(
                r".*_((\d*)" + re.escape(self.time_period_separator) + r"?(\d*)?).*"
            )
        return self._time_period_regex

    @property
    def timestamp_definitions(self):
        """dict: Definition of valid timestamp information and corresponding key values.

        This follows the CMIP standards where time strings must be one of the
        following: YYYY, YYYYMM, YYYYMMDD, YYYYMMDDHH or one of the previous combined
        with a hyphen e.g. YYYY-YYYY.

        Each key in the definitions dictionary is the length of the timestamp. Each
        value is itself a dictionary, with keys:

        - datetime_str: the string required to convert a timestamp of this length into a datetime using ``datetime.datetime.strptime``
        - generic_regexp: a regular expression which will match timestamps in this format
        - expected_timestep: a ``dateutil.relativedelta.relativedelta`` object which contains the expected timestep in files with this timestamp

        Examples
        --------
        >>> self.timestamp_definitions[len("2012")]["datetime_str"]
        "%Y"
        """
        if self._timestamp_definitions is None:
            self._timestamp_definitions = {}
            for key, value in self._known_timestamps.items():
                self._timestamp_definitions[key] = value
                generic_regexp = r"\d{" + "{}".format(key) + r"}"
                self._timestamp_definitions[key]["generic_regexp"] = generic_regexp
                hyphen_key = 2 * key + 1
                self._timestamp_definitions[hyphen_key] = {
                    "datetime_str": "{0}-{0}".format(value["datetime_str"]),
                    "expected_timestep": value["expected_timestep"],
                    "generic_regexp": "{0}-{0}".format(generic_regexp),
                }

        return self._timestamp_definitions

    @property
    def lon_dim(self):
        """:obj:`iris.coords.DimCoord` The longitude dimension of the data."""
        return self.cube.coord(self.lon_name)

    @property
    def lon_dim_number(self):
        """int: The index which corresponds to the longitude dimension.

        e.g. if longitude is the third dimension of the data, then
        ``self.lon_dim_number`` will be ``2`` (Python is zero-indexed).
        """
        return self.cube.coord_dims(self.lon_name)[0]

    @property
    def lat_dim(self):
        """:obj:`iris.coords.DimCoord` The latitude dimension of the data."""
        return self.cube.coord(self.lat_name)

    @property
    def lat_dim_number(self):
        """int: The index which corresponds to the latitude dimension.

        e.g. if latitude is the first dimension of the data, then
        ``self.lat_dim_number`` will be ``0`` (Python is zero-indexed).
        """
        return self.cube.coord_dims(self.lat_name)[0]

    @property
    def time_dim(self):
        """:obj:`iris.coords.DimCoord` The time dimension of the data."""
        return self.cube.coord(self.time_name)

    @property
    def time_dim_number(self):
        """int: The index which corresponds to the time dimension.

        e.g. if time is the first dimension of the data, then
        ``self.time_dim_number`` will be ``0`` (Python is zero-indexed).
        """
        return self.cube.coord_dims(self.time_name)[0]

    def _load_cube(self, filepath, constraint=None):
        self.cube = iris.load_cube(filepath, constraint=constraint)
        self._check_cube()

    def _check_cube(self):
        try:
            time_dim = self.time_dim
            gregorian = time_dim.units.calendar == "gregorian"
            year_zero = str(time_dim.units).startswith("days since 0-1-1")
        except CoordinateNotFoundError:
            gregorian = False
            year_zero = False

        if gregorian and year_zero:
            warn_msg = (
                "Your calendar is gregorian yet has units of 'days since 0-1-1'. "
                "We rectify this by removing all data before year 1 and changing the "
                "units to 'days since 1-1-1'. If you want other behaviour, you will "
                "need to use another package."
            )
            warnings.warn(warn_msg)
            self._adjust_gregorian_year_zero_units()

    def _adjust_gregorian_year_zero_units(self):
        year_zero_cube = self.cube.copy()
        year_zero_cube_time_dim = self.time_dim

        gregorian_year_zero_cube = (
            year_zero_cube_time_dim.units.calendar == "gregorian"
        ) and str(year_zero_cube_time_dim.units).startswith("days since 0-1-1")
        assert gregorian_year_zero_cube, "This function is not setup for other cases"

        new_unit_str = "days since 1-1-1"
        # converting with the new units means we're actually converting with the wrong
        # units, we use this variable to track how many years to shift back to get the
        # right time axis again
        new_units_shift = 1
        new_time_dim_unit = cf_units.Unit(
            new_unit_str, calendar=year_zero_cube_time_dim.units.calendar
        )

        tmp_time_dim = year_zero_cube_time_dim.copy()
        tmp_time_dim.units = new_time_dim_unit
        tmp_cube = iris.cube.Cube(year_zero_cube.data)
        for i, coord in enumerate(year_zero_cube.dim_coords):
            if coord.standard_name == "time":
                tmp_cube.add_dim_coord(tmp_time_dim, i)
            else:
                tmp_cube.add_dim_coord(coord, i)

        years_to_bin = 1
        first_valid_year = years_to_bin + new_units_shift

        def check_usable_data(cell):
            return first_valid_year <= cell.point.year

        usable_cube = tmp_cube.extract(iris.Constraint(time=check_usable_data))
        usable_data = usable_cube.data

        tmp_time_dim = usable_cube.coord(self.time_name)
        tmp_time = cftime.num2date(
            tmp_time_dim.points, new_unit_str, tmp_time_dim.units.calendar
        )
        # TODO: move to utils
        tmp_time = np.array([datetime(*v.timetuple()[:6]) for v in tmp_time])
        # undo the shift to new units
        usable_time = cf_units.date2num(
            tmp_time - relativedelta(years=new_units_shift),
            year_zero_cube_time_dim.units.name,
            year_zero_cube_time_dim.units.calendar,
        )
        usable_time_unit = cf_units.Unit(
            year_zero_cube_time_dim.units.name,
            calendar=year_zero_cube_time_dim.units.calendar,
        )
        usable_time_dim = iris.coords.DimCoord(
            usable_time,
            standard_name=year_zero_cube_time_dim.standard_name,
            long_name=year_zero_cube_time_dim.long_name,
            var_name=year_zero_cube_time_dim.var_name,
            units=usable_time_unit,
        )

        self.cube = iris.cube.Cube(usable_data)
        for i, coord in enumerate(usable_cube.dim_coords):
            if coord.standard_name == "time":
                self.cube.add_dim_coord(usable_time_dim, i)
            else:
                self.cube.add_dim_coord(coord, i)

        # hard coding as making this list dynamically is super hard as there's so many
        # edge cases to cover
        attributes_to_copy = [
            "attributes",
            "cell_methods",
            "units",
            "var_name",
            "standard_name",
            "name",
            "metadata",
            "long_name",
        ]
        for att in attributes_to_copy:
            setattr(self.cube, att, getattr(year_zero_cube, att))

    def load_data_from_path(self, filepath):
        """Load data from a path.

        If you are using the ``SCMCube`` class directly, this method simply loads the
        path into an iris cube which can be accessed through ``self.cube``.

        If implemented on a subclass of ``SCMCube``, this method should:

        - use ``self.get_load_data_from_identifiers_args_from_filepath`` to determine the suitable set of arguments to pass to ``self.load_data_from_identifiers`` from the filepath
        - load the data using ``self.load_data_from_identifiers`` as this method contains much better checks and helper components

        Parameters
        ----------
        filepath : str
            The filepath from which to load the data.
        """
        self._load_cube(filepath)

    def get_load_data_from_identifiers_args_from_filepath(self, filepath=None):
        """Get the set of identifiers to use to load data from a filepath.

        Parameters
        ----------
        filepath : str
            The filepath from which to load the data.

        Returns
        -------
        dict
            Set of arguments which can be passed to
            ``self.load_data_from_identifiers`` to load the data in the filepath.
        """
        raise NotImplementedError()

    def load_data_in_directory(self, directory=None):
        """Load data in a directory.

        The data is loaded into an iris cube which can be accessed through
        ``self.cube``.

        Initially, this method is intended to only be used to load data when it is
        saved in a number of different timeslice files e.g.:

        - tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc
        - tas_Amon_HadCM3_rcp45_r1i1p1_203101-203512.nc
        - tas_Amon_HadCM3_rcp45_r1i1p1_203601-203812.nc

        It is not intended to be used to load multiple different variables or
        non-continuous timeseries. These use cases could be added in future, but are
        not required yet so have not been included.

        Note that this function removes any attributes which aren't common between the
        loaded cubes. In general, we have found that this mainly means
        ``creation_date``, ``tracking_id`` and ``history`` are deleted. If unsure,
        please check.

        Parameters
        ----------
        directory : str
            Directory from which to load the data.

        Raises
        ------
        ValueError
            If the files in the directory are not from the same run (i.e. their filenames are not identical except for the timestamp) or if the files don't form a continuous timeseries.
        """
        self._load_and_concatenate_files_in_directory(directory)

    def _load_and_concatenate_files_in_directory(self, directory):
        self._check_data_names_in_same_directory(directory)

        # we use a loop here to make the most of finding missing data like
        # land-surface fraction and cellarea, something iris can't automatically do
        loaded_cubes = []
        for f in sorted(os.listdir(directory)):
            self.load_data_from_path(join(directory, f))
            loaded_cubes.append(self.cube)

        loaded_cubes = iris.cube.CubeList(loaded_cubes)

        unify_time_units(loaded_cubes)
        unify_lat_lon(loaded_cubes)
        iris.experimental.equalise_cubes.equalise_attributes(loaded_cubes)

        self.cube = loaded_cubes.concatenate_cube()

    def _check_data_names_in_same_directory(self, directory):
        found_files = sorted(os.listdir(directory))

        assertion_error_msg = (
            "Cannot join files in:\n"
            "{}\n"
            "Files found:\n"
            "- {}".format(directory, "\n- ".join(found_files))
        )

        base_regexp = self._get_timestamp_regex_from_filename(found_files[0])
        expected_timestep = self._get_expected_timestep_from_filename(found_files[0])

        file_timestamp_bits_prev = self._get_timestamp_bits_from_filename(
            found_files[0]
        )
        time_format = self.timestamp_definitions[
            len(file_timestamp_bits_prev["timestart_str"])
        ]["datetime_str"]
        for found_file in found_files[1:]:
            assert re.match(base_regexp, found_file), assertion_error_msg

            file_timestamp_bits = self._get_timestamp_bits_from_filename(found_file)
            end_time_prev = datetime.strptime(
                file_timestamp_bits_prev["timeend_str"], time_format
            )
            start_time = datetime.strptime(
                file_timestamp_bits["timestart_str"], time_format
            )

            assert (
                relativedelta(start_time, end_time_prev) == expected_timestep
            ), assertion_error_msg

            file_timestamp_bits_prev = file_timestamp_bits

    def _get_timestamp_regex_from_filename(self, filename):
        timestamp_bits = self._get_timestamp_bits_from_filename(filename)
        timestamp_str = timestamp_bits["timestamp_str"]
        return filename.replace(
            timestamp_str,
            self.timestamp_definitions[len(timestamp_str)]["generic_regexp"],
        )

    def _get_expected_timestep_from_filename(self, filename):
        timestamp_bits = self._get_timestamp_bits_from_filename(filename)
        timestamp_str = timestamp_bits["timestamp_str"]
        return self.timestamp_definitions[len(timestamp_str)]["expected_timestep"]

    def _get_timestamp_bits_from_filename(self, filename):
        regex_matches = re.match(self.time_period_regex, filename)
        timestamp_str = regex_matches.group(1)
        self._check_time_period_valid(timestamp_str)
        start_time = regex_matches.group(2)
        end_time = regex_matches.group(3)
        timestep = self.timestamp_definitions[len(timestamp_str)]["expected_timestep"]
        return {
            "timestamp_str": timestamp_str,
            "timestart_str": start_time,
            "timeend_str": end_time,
            "expected_timestep": timestep,
        }

    def load_data_from_identifiers(self, **kwargs):
        """Load data using key identifiers.

        The identifiers are used to determine the path of the file to load. The file
        is then loaded into an iris cube which can be accessed through ``self.cube``.

        Parameters
        ----------
        **kwargs
            Arguments which can then be processed by
            ``self.get_filepath_from_load_data_from_identifiers_args`` and
            ``self.get_variable_constraint_from_load_data_from_identifiers_args`` to determine the full
            filepath of the file to load and the variable constraint to use.
        """
        with warnings.catch_warnings(record=True) as w:
            self._load_cube(
                self.get_filepath_from_load_data_from_identifiers_args(**kwargs),
                self.get_variable_constraint_from_load_data_from_identifiers_args(
                    **kwargs
                ),
            )

        if w:
            self._process_load_data_from_identifiers_warnings(w)

    def _process_load_data_from_identifiers_warnings(self, w):
        area_cell_warn = "Missing CF-netCDF measure variable 'areacella'"
        for warn in w:
            if area_cell_warn in str(warn.message):
                try:
                    self._add_areacella_measure()
                except Exception:
                    custom_warn = (
                        "Tried to add areacella cube, failed as shown:\n"
                        + traceback.format_exc()
                    )
                    warnings.warn(custom_warn)
                    warn_message = "\n\nareacella warning:\n" + str(warn.message)
                    warnings.warn(warn_message)
            else:
                warnings.warn(warn.message)

    def _add_areacella_measure(self):
        areacella_cube = self.get_metadata_cube(self.areacella_var).cube
        areacella_measure = iris.coords.CellMeasure(
            areacella_cube.data,
            standard_name=areacella_cube.standard_name,
            long_name=areacella_cube.long_name,
            var_name=areacella_cube.var_name,
            units=areacella_cube.units,
            attributes=areacella_cube.attributes,
            measure="area",
        )
        self.cube.add_cell_measure(
            areacella_measure, data_dims=[self.lat_dim_number, self.lon_dim_number]
        )

    def get_filepath_from_load_data_from_identifiers_args(self, **kwargs):
        """Get the full filepath of the data to load from the arguments passed to ``self.load_data_from_identifiers``.

        This function should, in most cases, call ``self._get_data_directory`` and
        ``self._get_data_filename``.

        Parameters
        ----------
        **kwargs
            Arguments, initially passed to ``self.load_data_from_identifiers`` from which the full
            filepath of the file to load should be determined.

        Returns
        -------
        str
            The full filepath (path and name) of the file to load.
        """
        raise NotImplementedError()

    def get_variable_constraint_from_load_data_from_identifiers_args(self, **kwargs):
        """Get the iris variable constraint to use when loading data with ``self.load_data_from_identifiers``.

        Parameters
        ----------
        **kwargs
            Arguments, initially passed to ``self.load_data_from_identifiers`` from which the full
            filepath of the file to load should be determined.

        Returns
        -------
        :obj:`iris.Constraint`
            constraint to use which ensures that only the variable of interest is loaded.
        """
        raise NotImplementedError()

    def _get_data_directory(self):
        """Get the path to a data file from self's attributes.

        This can take multiple forms, it may just return a previously set
        filepath attribute or it could combine a number of different metadata
        elements (e.g. model name, experiment name) to create the data path.

        Returns
        -------
        str
            path to the data file from which this cube has been/will be loaded
        """
        raise NotImplementedError()

    def _get_data_filename(self):
        """Get the name of a data file from self's attributes.

        This can take multiple forms, it may just return a previously set
        filename attribute or it could combine a number of different metadata
        elements (e.g. model name, experiment name) to create the data name.

        Returns
        -------
        str
            name of the data file from which this cube has been/will be loaded.
        """
        raise NotImplementedError()

    def get_metadata_cube(self, metadata_variable):
        """Load a metadata cube from self's attributes.

        Parameters
        ----------
        metadata_variable : str
            the name of the metadata variable to get, as it appears in the filename.

        Returns
        -------
        :obj:`type(self)`
            instance of self which has been loaded from the file containing the metadata variable of interest.
        """
        load_args = self._get_metadata_load_arguments(metadata_variable)

        metadata_cube = type(self)()
        metadata_cube.load_data_from_identifiers(**load_args)

        return metadata_cube

    def _get_metadata_load_arguments(self, metadata_variable):
        """Get the arguments to load a metadata file from self's attributes.

        This can take multiple forms, it may just return a previously set
        metada_filename attribute or it could combine a number of different
        metadata elements (e.g. model name, experiment name) to create the
        metadata filename.

        Parameters
        ----------
        metadata_variable : str
            the name of the metadata variable to get, as it appears in the filename.

        Returns
        -------
        dict
            dictionary containing all the arguments to pass to ``self.load_data_from_identifiers``
            required to load the desired metadata cube.
        """
        raise NotImplementedError()

    def get_scm_timeseries(
        self, sftlf_cube=None, land_mask_threshold=50, areacella_scmcube=None
    ):
        """Get SCM relevant timeseries from ``self``.

        Parameters
        ----------
        sftlf_cube : :obj:`SCMCube`, optional
            land surface fraction data which is used to determine whether a given
            gridbox is land or ocean. If ``None``, we try to load the land surface fraction automatically.

        land_mask_threshold : float, optional
            if the surface land fraction in a grid box is greater than
            ``land_mask_threshold``, it is considered to be a land grid box.

        areacella_scmcube : :obj:`SCMCube`, optional
            cell area data which is used to take the latitude-longitude mean of the
            cube's data. If ``None``, we try to load this data automatically and if
            that fails we fall back onto ``iris.analysis.cartography.area_weights``.

        Returns
        -------
        :obj:`openscm.io.ScmDataFrame`
            An OpenSCM DataFrame instance with the data in the ``data`` attribute and
            metadata in the ``metadata`` attribute.
        """
        scm_timeseries_cubes = self.get_scm_timeseries_cubes(
            sftlf_cube=sftlf_cube,
            land_mask_threshold=land_mask_threshold,
            areacella_scmcube=areacella_scmcube,
        )

        return self._convert_scm_timeseries_cubes_to_openscmdata(scm_timeseries_cubes)

    def get_scm_timeseries_cubes(
        self, sftlf_cube=None, land_mask_threshold=50, areacella_scmcube=None
    ):
        """Get SCM relevant cubes from the ``self``.

        Parameters
        ----------
        sftlf_cube : :obj:`SCMCube`, optional
            land surface fraction data which is used to determine whether a given
            gridbox is land or ocean. If ``None``, we try to load the land surface fraction automatically.

        land_mask_threshold : float, optional
            if the surface land fraction in a grid box is greater than
            ``land_mask_threshold``, it is considered to be a land grid box.

        areacella_scmcube : :obj:`SCMCube`, optional
            cell area data which is used to take the latitude-longitude mean of the
            cube's data. If ``None``, we try to load this data automatically and if
            that fails we fall back onto ``iris.analysis.cartography.area_weights``.

        Returns
        -------
        dict
            Cubes, with latitude-longitude mean data as appropriate for each of the
            SCM relevant regions.
        """
        area_weights = self._get_area_weights(areacella_scmcube=areacella_scmcube)
        scm_cubes = self.get_scm_cubes(
            sftlf_cube=sftlf_cube, land_mask_threshold=land_mask_threshold
        )
        return {
            k: take_lat_lon_mean(scm_cube, area_weights)
            for k, scm_cube in scm_cubes.items()
        }

    def get_scm_cubes(self, sftlf_cube=None, land_mask_threshold=50):
        """Get SCM relevant cubes from the ``self``.

        Parameters
        ----------
        sftlf_cube : :obj:`SCMCube`, optional
            land surface fraction data which is used to determine whether a given
            gridbox is land or ocean. If ``None``, we try to load the land surface fraction automatically.

        land_mask_threshold : float, optional
            if the surface land fraction in a grid box is greater than
            ``land_mask_threshold``, it is considered to be a land grid box.

        Returns
        -------
        dict
            Cubes, with data masked as appropriate for each of the SCM relevant
            regions.
        """
        scm_masks = self._get_scm_masks(
            sftlf_cube=sftlf_cube, land_mask_threshold=land_mask_threshold
        )

        return {k: apply_mask(self, mask) for k, mask in scm_masks.items()}

    def _get_scm_masks(self, sftlf_cube=None, land_mask_threshold=50):
        """Get the scm masks.

        Returns
        -------
        dict
        """
        nh_mask = self._get_nh_mask()
        hemisphere_masks = {
            "World": np.full(nh_mask.shape, False),
            "World|Northern Hemisphere": nh_mask,
            "World|Southern Hemisphere": ~nh_mask,
        }

        try:
            land_mask = self._get_land_mask(
                sftlf_cube=sftlf_cube, land_mask_threshold=land_mask_threshold
            )
            land_masks = {
                "World|Northern Hemisphere|Land": np.logical_or(nh_mask, land_mask),
                "World|Southern Hemisphere|Land": np.logical_or(~nh_mask, land_mask),
                "World|Northern Hemisphere|Ocean": np.logical_or(nh_mask, ~land_mask),
                "World|Southern Hemisphere|Ocean": np.logical_or(~nh_mask, ~land_mask),
                "World|Land": land_mask,
                "World|Ocean": ~land_mask,
            }

            return {**hemisphere_masks, **land_masks}

        except OSError:
            warn_msg = (
                "Land surface fraction (sftlf) data not available, only returning "
                "global and hemispheric masks."
            )
            warnings.warn(warn_msg)
            return hemisphere_masks

    def _get_land_mask(self, sftlf_cube=None, land_mask_threshold=50):
        """Get the land mask.

        Returns
        -------
        np.ndarray
        """
        if sftlf_cube is None:
            sftlf_cube = self.get_metadata_cube(self.sftlf_var)

        if not isinstance(sftlf_cube, SCMCube):
            raise TypeError("sftlf_cube must be an SCMCube instance")

        sftlf_data = sftlf_cube.cube.data

        land_mask = np.where(
            sftlf_data > land_mask_threshold,
            False,  # where it's land, return False i.e. don't mask
            True,  # otherwise True
        )

        return self._broadcast_onto_self_lat_lon_grid(land_mask)

    def _broadcast_onto_self_lat_lon_grid(self, array_in):
        """Broadcast an array onto the latitude-longitude grid of ``self``.

        Here, broadcasting means taking the array and 'duplicating' it so that it
        has the same number of dimensions as the cube's underlying data. For example,
        if our cube has a time dimension of length 3, a latitude dimension of length 4
        and a longitude dimension of length 2 then if we are given in a 4x2 array, we
        broadcast this onto a 3x4x2 array where each slice in the broadcasted array's
        time dimension is identical to the input array.
        """
        lat_length = len(self.lat_dim.points)
        lon_length = len(self.lon_dim.points)

        dim_order = [self.lat_dim_number, self.lon_dim_number]
        base_shape = (lat_length, lon_length)
        if array_in.shape != base_shape:
            array_in = np.transpose(array_in)

        shape_assert_msg = (
            "the sftlf_cube data must be the same shape as the "
            "cube's longitude-latitude grid"
        )
        assert array_in.shape == base_shape, shape_assert_msg

        return broadcast_to_shape(array_in, self.cube.shape, dim_order)

    def _get_nh_mask(self):
        mask_nh_lat = np.array(
            [cell < 0 for cell in self.cube.coord(self.lat_name).cells()]
        )
        mask_all_lon = np.full(self.cube.coord(self.lon_name).points.shape, False)

        # Here we make a grid which we can use as a mask. We have to use all
        # of these nots so that our product (which uses AND logic) gives us
        # False in the NH and True in the SH (another way to think of this is
        # that we have to flip everything so False goes to True and True goes
        # to False, do all our operations with AND logic, then flip everything
        # back).
        mask_nh = ~np.outer(~mask_nh_lat, ~mask_all_lon)

        return self._broadcast_onto_self_lat_lon_grid(mask_nh)

    def _get_area_weights(self, areacella_scmcube=None):
        use_self_area_weights = True
        if areacella_scmcube is None:
            with warnings.catch_warnings(record=True) as w:
                areacella_scmcube = self._get_areacella_scmcube()
            if w:
                use_self_area_weights = False
                for warn in w:
                    warnings.warn(warn.message)

        if use_self_area_weights:
            try:
                areacella_cube = areacella_scmcube.cube
                return self._broadcast_onto_self_lat_lon_grid(areacella_cube.data)
            except AssertionError as exc:
                warnings.warn(str(exc))

        warnings.warn(
            "Couldn't find/use areacella_cube, falling back to iris.analysis.cartography.area_weights"
        )
        return iris.analysis.cartography.area_weights(self.cube)

    def _get_areacella_scmcube(self):
        try:
            areacella_scmcube = self.get_metadata_cube(self.areacella_var)
            if not isinstance(areacella_scmcube.cube, iris.cube.Cube):
                warnings.warn(
                    "areacella cube which was found has cube attribute which isn't an iris cube"
                )
            else:
                return areacella_scmcube
        except iris.exceptions.ConstraintMismatchError as exc:
            warnings.warn(str(exc))
        except AttributeError as exc:
            warnings.warn(str(exc))
        except OSError as exc:
            warnings.warn(str(exc))
        except NotImplementedError as exc:
            warnings.warn(str(exc))

    def _convert_scm_timeseries_cubes_to_openscmdata(
        self, scm_timeseries_cubes, out_calendar=None
    ):
        # could probably just use iris.pandas.to_series() here..?
        data = {k: get_cube_timeseries_data(v) for k, v in scm_timeseries_cubes.items()}

        time_index, out_calendar = self._get_openscmdata_time_axis_and_calendar(
            scm_timeseries_cubes, out_calendar=out_calendar
        )

        climate_model, scenario = self._get_climate_model_scenario()

        out_df = pd.DataFrame(data, index=time_index)
        out_df.columns = pd.MultiIndex.from_product(
            [
                [self.cube.standard_name],
                [self.cube.units.name],
                out_df.columns.tolist(),
                [climate_model],
                [scenario],
                ["unspecified"],
            ],
            names=["variable", "unit", "region", "climate_model", "scenario", "model"],
        )
        out_df = out_df.unstack().reset_index().rename({0: "value"}, axis="columns")
        output = ScmDataFrame(out_df)
        try:
            output.metadata["calendar"] = out_calendar
        except AttributeError:
            output.metadata = {"calendar": out_calendar}

        return output

    def _get_climate_model_scenario(self):
        try:
            climate_model = self.model
            scenario = "_".join([self.activity, self.experiment, self.ensemble_member])
        except AttributeError:
            warn_msg = (
                "Could not determine appropriate climate_model scenario combination, "
                "filling with 'unspecified'"
            )
            warnings.warn(warn_msg)
            climate_model = "unspecified"
            scenario = "unspecified"

        return climate_model, scenario

    def _get_openscmdata_time_axis_and_calendar(
        self, scm_timeseries_cubes, out_calendar
    ):
        if out_calendar is None:
            out_calendar = self.cube.coords("time")[0].units.calendar

        time_axes = [
            get_scm_cube_time_axis_in_calendar(scm_cube, out_calendar)
            for scm_cube in scm_timeseries_cubes.values()
        ]
        assert_all_time_axes_same(time_axes)
        time_axis = time_axes[0]

        if isinstance(time_axis[0], cftime.datetime):
            time_axis = np.array([parser.parse(x.strftime()) for x in time_axis])
        elif isinstance(time_axis[0], datetime):
            pass

        # As we sometimes have to deal with long timeseries, we force the index to be
        # pd.Index and not pd.DatetimeIndex. We can't use DatetimeIndex because of a
        # pandas limitation, see
        # http://pandas-docs.github.io/pandas-docs-travis/timeseries.html#timestamp-limitations
        return pd.Index(time_axis, dtype="object", name="time"), out_calendar

    def _check_time_period_valid(self, time_period_str):
        """Check that a time_period identifier string is valid.

        Parameters
        ----------
        time_period_str : str
            Time period string to check.

        Raises
        ------
        ValueError
            If the time period is not in a valid format.
        """
        if "_" in time_period_str:
            self._raise_time_period_invalid_error(time_period_str)

        if "-" in time_period_str:
            dates = time_period_str.split("-")
            if (len(dates) != 2) or (len(dates[0]) != len(dates[1])):
                self._raise_time_period_invalid_error(time_period_str)
        else:
            dates = [time_period_str]

        try:
            time_format = self.timestamp_definitions[len(dates[0])]["datetime_str"]
        except KeyError:
            self._raise_time_period_invalid_error(time_period_str)
        for date in dates:
            try:
                datetime.strptime(date, time_format)
            except ValueError:
                self._raise_time_period_invalid_error(time_period_str)

        if len(dates) == 2:
            start_date = datetime.strptime(dates[0], time_format)
            end_date = datetime.strptime(dates[1], time_format)
            if not start_date < end_date:
                self._raise_time_period_invalid_error(time_period_str)

    def _raise_time_period_invalid_error(self, time_period_str):
        message = "Your time_period indicator ({}) does not look right".format(
            time_period_str
        )
        raise ValueError(message)


class _CMIPCube(SCMCube):
    def load_data_from_path(self, filepath):
        """Load data from a path.

        Parameters
        ----------
        filepath : str
            The filepath from which to load the data.
        """
        load_data_from_identifiers_args = self.get_load_data_from_identifiers_args_from_filepath(
            filepath
        )
        self.load_data_from_identifiers(**load_data_from_identifiers_args)

    def _load_and_concatenate_files_in_directory(self, directory):
        super()._load_and_concatenate_files_in_directory(directory)
        self._add_time_period_from_files_in_directory(directory)

    def _add_time_period_from_files_in_directory(self, directory):
        self._check_data_names_in_same_directory(directory)

        loaded_files = sorted(os.listdir(directory))
        strt = self._get_timestamp_bits_from_filename(loaded_files[0])["timestart_str"]
        end = self._get_timestamp_bits_from_filename(loaded_files[-1])["timeend_str"]
        self.time_period = self.time_period_separator.join([strt, end])

    def get_load_data_from_identifiers_args_from_filepath(self, filepath):
        """Get the set of identifiers to use to load data from a filepath.

        Parameters
        ----------
        filepath : str
            The filepath from which to load the data.

        Returns
        -------
        dict
            Set of arguments which can be passed to
            ``self.load_data_from_identifiers`` to load the data in the filepath.
        """
        path_ids = self.process_path(dirname(filepath))
        name_ids = self.process_filename(basename(filepath))
        for key, value in path_ids.items():
            if (key in name_ids) and (value != name_ids[key]):
                error_msg = (
                    "Path and filename do not agree:\n"
                    "    - path {0}: {1}\n"
                    "    - filename {0}: {2}\n".format(key, value, name_ids[key])
                )
                raise ValueError(error_msg)

        return {**path_ids, **name_ids}

    def _raise_path_error(self, path):
        raise ValueError("Path does not look right: {}".format(path))

    def _raise_filename_error(self, filename):
        raise ValueError("Filename does not look right: {}".format(filename))


class MarbleCMIP5Cube(_CMIPCube):
    """Subclass of ``SCMCube`` which can be used with the ``cmip5`` directory on marble.

    This directory structure is very similar, but not quite identical, to the
    recommended CMIP5 directory structure described in section 3.1 of the
    `CMIP5 Data Reference Syntax <https://cmip.llnl.gov/cmip5/docs/cmip5_data_reference_syntax_v1-00_clean.pdf>`_.
    """

    def process_filename(self, filename):
        """Cut a filename into its identifiers

        Parameters
        ----------
        filename : str
            The filename to process. Filename here means just the filename, no path
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input filename
        """
        filename_bits = filename.split("_")
        if len(filename_bits) == 6:
            time_period, file_ext = splitext(filename_bits[-1])
            ensemble_member = filename_bits[-2]
        elif len(filename_bits) == 5:
            time_period = None
            ensemble_member, file_ext = splitext(filename_bits[-1])
        else:
            self._raise_filename_error(filename)

        if not file_ext:
            self._raise_filename_error(filename)

        return {
            "variable_name": filename_bits[0],
            "modeling_realm": filename_bits[1],
            "model": filename_bits[2],
            "experiment": filename_bits[3],
            "ensemble_member": ensemble_member,
            "time_period": time_period,
            "file_ext": file_ext,
        }

    def process_path(self, path):
        """Cut a path into its identifiers

        Parameters
        ----------
        path : str
            The path to process. Path here means just the path, no filename
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input path
        """
        path = path[:-1] if path.endswith(os.sep) else path
        dirpath_bits = path.split(os.sep)
        if (len(dirpath_bits) < 6) or any(["_" in d for d in dirpath_bits[-6:]]):
            self._raise_path_error(path)

        root_dir = os.sep.join(dirpath_bits[:-6])
        if not root_dir:
            root_dir = "."

        return {
            "root_dir": root_dir,
            "activity": dirpath_bits[-6],
            "variable_name": dirpath_bits[-3],
            "modeling_realm": dirpath_bits[-4],
            "model": dirpath_bits[-2],
            "experiment": dirpath_bits[-5],
            "ensemble_member": dirpath_bits[-1],
        }

    def get_filepath_from_load_data_from_identifiers_args(
        self,
        root_dir=".",
        activity="activity",
        experiment="experiment",
        modeling_realm="modeling-realm",
        variable_name="variable-name",
        model="model",
        ensemble_member="ensemble-member",
        time_period=None,
        file_ext=".nc",
    ):
        """Get the full filepath of the data to load from the arguments passed to ``self.load_data_from_identifiers``.

        Full details about the identifiers are given in Section 2 of the
        `CMIP5 Data Reference Syntax <https://cmip.llnl.gov/cmip5/docs/cmip5_data_reference_syntax_v1-00_clean.pdf>`_.

        Parameters
        ----------
        root_dir : str, optional
            The root directory of the database i.e. where the cube should start its
            path from e.g. ``/home/users/usertim/cmip5_25x25``.

        activity : str, optional
            The activity for which we want to load data e.g. ``cmip5``.

        experiment : str, optional
            The experiment for which we want to load data e.g. ``1pctCO2``.

        modeling_realm : str, optional
            The modeling_realm for which we want to load data e.g. ``Amon``.

        variable_name : str, optional
            The variable for which we want to load data e.g. ``variable_name``.

        model : str, optional
            The model for which we want to load data ``CanESM2``.

        ensemble_member : str, optional
            The ensemble member for which we want to load data ``r1i1p1``.

        time_period : str, optional
            The time period for which we want to load data e.g. ``1850-2000``.
            If ``None``, this information isn't included in the filename which is
            useful for loading metadata files which don't have a relevant time period.

        file_ext : str, optional
            The file extension of the data file we want to load e.g. ``.nc``.

        Returns
        -------
        str
            The full filepath (path and name) of the file to load.
        """
        inargs = locals()
        del inargs["self"]
        # if the step above ever gets more complicated, use the solution here
        # http://kbyanc.blogspot.com/2007/07/python-aggregating-function-arguments.html

        for name, value in inargs.items():
            setattr(self, name, value)

        return join(self._get_data_directory(), self._get_data_filename())

    def _get_data_directory(self):
        return join(
            self.root_dir,
            self.activity,
            self.experiment,
            self.modeling_realm,
            self.variable_name,
            self.model,
            self.ensemble_member,
        )

    def _get_data_filename(self):
        bits_to_join = [
            self.variable_name,
            self.modeling_realm,
            self.model,
            self.experiment,
            self.ensemble_member,
        ]
        if self.time_period is not None:
            bits_to_join.append(self.time_period)

        return "_".join(bits_to_join) + self.file_ext

    def get_variable_constraint_from_load_data_from_identifiers_args(
        self, variable_name="tas", **kwargs
    ):
        """Get the iris variable constraint to use when loading data with ``self.load_data_from_identifiers``.

        Parameters
        ----------
        root_dir : str, optional
            The root directory of the database i.e. where the cube should start its
            path from e.g. ``/home/users/usertim/cmip5_25x25``.

        activity : str, optional
            The activity for which we want to load data.

        experiment : str, optional
            The experiment for which we want to load data.

        modeling_realm : str, optional
            The modeling_realm for which we want to load data.

        variable_name : str, optional
            The variable for which we want to load data.

        model : str, optional
            The model for which we want to load data.

        ensemble_member : str, optional
            The ensemble member for which we want to load data.

        time_period : str, optional
            The time period for which we want to load data. If ``None``, this
            information isn't included in the filename which is useful for loading
            metadata files which don't have a relevant time period.

        file_ext : str, optional
            The file extension of the data file we want to load.

        Returns
        -------
        :obj:`iris.Constraint`
            constraint to use which ensures that only the variable of interest is loaded.
        """
        # thank you Duncan!!
        # https://github.com/SciTools/iris/issues/2107#issuecomment-246644471
        return iris.Constraint(
            cube_func=(lambda c: c.var_name == np.str(variable_name))
        )

    def _get_metadata_load_arguments(self, metadata_variable):
        return {
            "root_dir": self.root_dir,
            "activity": self.activity,
            "experiment": self.experiment,
            "modeling_realm": "fx",
            "variable_name": metadata_variable,
            "model": self.model,
            "ensemble_member": "r0i0p0",
            "time_period": None,
            "file_ext": self.file_ext,
        }


class CMIP6Input4MIPsCube(_CMIPCube):
    """Subclass of ``SCMCube`` which can be used with CMIP6 input4MIPs data

    The data must match the CMIP6 Forcing Datasets Summary, specifically the
    `Forcing Dataset Specifications <https://docs.google.com/document/d/1pU9IiJvPJwRvIgVaSDdJ4O0Jeorv_2ekEtted34K9cA/edit#heading=h.cn9f7982ycw6>`_.
    """

    def process_filename(self, filename):
        """Cut a filename into its identifiers

        Parameters
        ----------
        filename : str
            The filename to process. Filename here means just the filename, no path
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input filename
        """
        filename_bits = filename.split("_")
        if len(filename_bits) == 7:
            time_range, file_ext = splitext(filename_bits[-1])
            grid_label = filename_bits[-2]
        elif len(filename_bits) == 6:
            time_range = None
            grid_label, file_ext = splitext(filename_bits[-1])
        else:
            self._raise_filename_error(filename)

        if not file_ext:
            self._raise_filename_error(filename)

        return {
            "variable_id": filename_bits[0],
            "activity_id": filename_bits[1],
            "dataset_category": filename_bits[2],
            "target_mip": filename_bits[3],
            "source_id": filename_bits[4],
            "grid_label": grid_label,
            "time_range": time_range,
            "file_ext": file_ext,
        }

    def process_path(self, path):
        """Cut a path into its identifiers

        Parameters
        ----------
        path : str
            The path to process. Path here means just the path, no filename
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input path
        """
        path = path[:-1] if path.endswith(os.sep) else path
        dirpath_bits = path.split(os.sep)
        if (len(dirpath_bits) < 10) or any(["_" in d for d in dirpath_bits[-10:]]):
            self._raise_path_error(path)

        root_dir = os.sep.join(dirpath_bits[:-10])
        if not root_dir:
            root_dir = "."

        return {
            "root_dir": root_dir,
            "activity_id": dirpath_bits[-10],
            "mip_era": dirpath_bits[-9],
            "target_mip": dirpath_bits[-8],
            "institution_id": dirpath_bits[-7],
            "source_id": dirpath_bits[-6],
            "realm": dirpath_bits[-5],
            "frequency": dirpath_bits[-4],
            "variable_id": dirpath_bits[-3],
            "grid_label": dirpath_bits[-2],
            "version": dirpath_bits[-1],
        }

    def get_filepath_from_load_data_from_identifiers_args(
        self,
        root_dir=".",
        activity_id="activity-id",
        mip_era="mip-era",
        target_mip="target-mip",
        institution_id="institution-id",
        source_id="source-id-including-institution-id",
        realm="realm",
        frequency="frequency",
        variable_id="variable-id",
        grid_label="grid-label",
        version="version",
        dataset_category="dataset-category",
        time_range=None,
        file_ext="file-ext",
    ):
        """Get the full filepath of the data to load from the arguments passed to ``self.load_data_from_identifiers``.

        Full details about the meaning of the identifiers are given in the
        `Forcing Dataset Specifications <https://docs.google.com/document/d/1pU9IiJvPJwRvIgVaSDdJ4O0Jeorv_2ekEtted34K9cA/edit#heading=h.cn9f7982ycw6>`_.

        Parameters
        ----------
        root_dir : str, optional
            The root directory of the database i.e. where the cube should start its
            path from e.g. ``/home/users/usertim/cmip5_25x25``.

        activity_id : str, optional
            The activity_id for which we want to load data. For these cubes, will
            almost always be ``input4MIPs``.

        mip_era : str, optional
            The mip_era for which we want to load data e.g. ``CMIP6``.

        target_mip : str, optional
            The target_mip for which we want to load data e.g. ``ScenarioMIP``.

        institution_id : str, optional
            The institution_id for which we want to load data e.g. ``UoM``.

        source_id : str, optional
            The source_id for which we want to load data e.g.
            ``UoM-REMIND-MAGPIE-ssp585-1-2-0``. This must include the institution_id.

        realm : str, optional
            The realm for which we want to load data e.g. ``atmos``.

        frequency : str, optional
            The frequency for which we want to load data e.g. ``yr``.

        variable_id : str, optional
            The variable_id for which we want to load data e.g.
            ``mole-fraction-of-carbon-dioxide-in-air``.

        grid_label : str, optional
            The grid_label for which we want to load data e.g. ``gr1-GMNHSH``.

        version : str, optional
            The version for which we want to load data e.g. ``v20180427``.

        dataset_category : str, optional
            The dataset_category for which we want to load data e.g.
            ``GHGConcentrations``.

        time_range : str, optional
            The time range for which we want to load data e.g. ``2005-2100``. If
            ``None``, this information isn't included in the filename which is useful
            for loading metadata files which don't have a relevant time period.

        file_ext : str, optional
            The file extension of the data file we want to load e.g. ``.nc``.

        Returns
        -------
        str
            The full filepath (path and name) of the file to load.
        """
        inargs = locals()
        del inargs["self"]
        # if the step above ever gets more complicated, use the solution here
        # http://kbyanc.blogspot.com/2007/07/python-aggregating-function-arguments.html

        for name, value in inargs.items():
            setattr(self, name, value)

        # TODO: do time indicator/frequency checks too and make a new method for checks so can be reused by different methods
        self._check_self_consistency()

        return join(self._get_data_directory(), self._get_data_filename())

    def get_variable_constraint_from_load_data_from_identifiers_args(
        self, variable_id="tas", **kwargs
    ):
        """Get the iris variable constraint to use when loading data with ``self.load_data_from_identifiers``.

        Parameters
        ----------
        root_dir : str, optional
            The root directory of the database i.e. where the cube should start its
            path from e.g. ``/home/users/usertim/cmip5_25x25``.

        activity_id : str, optional
            The activity_id for which we want to load data. For these cubes, will
            almost always be "input4MIPs".

        mip_era : str, optional
            The mip_era for which we want to load data.

        target_mip : str, optional
            The target_mip for which we want to load data.

        institution_id : str, optional
            The institution_id for which we want to load data.

        source_id : str, optional
            The source_id for which we want to load data. This must include the version and the institution_id.

        realm : str, optional
            The realm for which we want to load data.

        frequency : str, optional
            The frequency for which we want to load data.

        variable_id : str, optional
            The variable_id for which we want to load data.

        grid_label : str, optional
            The grid_label for which we want to load data.

        version : str, optional
            The version for which we want to load data.

        dataset_category : str, optional
            The dataset_category for which we want to load data.

        time_range : str, optional
            The time range for which we want to load data. If ``None``, this
            information isn't included in the filename which is useful for loading
            metadata files which don't have a relevant time period.

        file_ext : str, optional
            The file extension of the data file we want to load.

        Returns
        -------
        :obj:`iris.Constraint`
            constraint to use which ensures that only the variable of interest is loaded.
        """
        # thank you Duncan!!
        # https://github.com/SciTools/iris/issues/2107#issuecomment-246644471
        return iris.Constraint(
            cube_func=(lambda c: c.var_name == np.str(variable_id.replace("-", "_")))
        )

    def _check_self_consistency(self):
        assert (
            self.institution_id in self.source_id
        ), "source_id must contain institution_id"

    def _get_metadata_load_arguments(self, metadata_variable):
        return {
            "root_dir": self.root_dir,
            "activity_id": self.activity_id,
            "mip_era": self.mip_era,
            "target_mip": self.target_mip,
            "institution_id": self.institution_id,
            "source_id": self.source_id,
            "realm": self.realm,
            "frequency": "fx",
            "variable_id": metadata_variable,
            "grid_label": self.grid_label,
            "version": self.version,
            "dataset_category": self.dataset_category,
            "time_range": None,
            "file_ext": self.file_ext,
        }

    def _get_data_filename(self):
        bits_to_join = [
            self.variable_id,
            self.activity_id,
            self.dataset_category,
            self.target_mip,
            self.source_id,
            self.grid_label,
        ]
        if self.time_range is not None:
            bits_to_join.append(self.time_range)

        return "_".join(bits_to_join) + self.file_ext

    def _get_data_directory(self):
        return join(
            self.root_dir,
            self.activity_id,
            self.mip_era,
            self.target_mip,
            self.institution_id,
            self.source_id,
            self.realm,
            self.frequency,
            self.variable_id,
            self.grid_label,
            self.version,
        )


class CMIP6OutputCube(_CMIPCube):
    """Subclass of ``SCMCube`` which can be used with CMIP6 model output data

    The data must match the CMIP6 data reference syntax as specified in the 'File name
    template' and 'Directory structure template' sections of the
    `CMIP6 Data Reference Syntax <https://goo.gl/v1drZl>`_.
    """

    def process_filename(self, filename):
        """Cut a filename into its identifiers

        Parameters
        ----------
        filename : str
            The filename to process. Filename here means just the filename, no path
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input filename
        """
        filename_bits = filename.split("_")
        if len(filename_bits) == 7:
            time_range, file_ext = splitext(filename_bits[-1])
            grid_label = filename_bits[-2]
        elif len(filename_bits) == 6:
            time_range = None
            grid_label, file_ext = splitext(filename_bits[-1])
        else:
            self._raise_filename_error(filename)

        if not file_ext:
            self._raise_filename_error(filename)

        return {
            "variable_id": filename_bits[0],
            "table_id": filename_bits[1],
            "source_id": filename_bits[2],
            "experiment_id": filename_bits[3],
            "member_id": filename_bits[4],
            "grid_label": grid_label,
            "time_range": time_range,
            "file_ext": file_ext,
        }

    def process_path(self, path):
        """Cut a path into its identifiers

        Parameters
        ----------
        path : str
            The path to process. Path here means just the path, no filename
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input path
        """
        path = path[:-1] if path.endswith(os.sep) else path
        dirpath_bits = path.split(os.sep)
        if (len(dirpath_bits) < 10) or any(["_" in d for d in dirpath_bits[-10:]]):
            self._raise_path_error(path)

        root_dir = os.sep.join(dirpath_bits[:-10])
        if not root_dir:
            root_dir = "."

        return {
            "root_dir": root_dir,
            "mip_era": dirpath_bits[-10],
            "activity_id": dirpath_bits[-9],
            "institution_id": dirpath_bits[-8],
            "source_id": dirpath_bits[-7],
            "experiment_id": dirpath_bits[-6],
            "member_id": dirpath_bits[-5],
            "table_id": dirpath_bits[-4],
            "variable_id": dirpath_bits[-3],
            "grid_label": dirpath_bits[-2],
            "version": dirpath_bits[-1],
        }

    def get_filepath_from_load_data_from_identifiers_args(
        self,
        root_dir=".",
        mip_era="mip-era",
        activity_id="activity-id",
        institution_id="institution-id",
        source_id="source-id",
        experiment_id="experiment-id",
        member_id="member-id",
        table_id="table-id",
        variable_id="variable-id",
        grid_label="grid-label",
        version="version",
        time_range=None,
        file_ext="file-ext",
    ):
        """Get the full filepath of the data to load from the arguments passed to ``self.load_data_from_identifiers``.

        Full details about the meaning of each identifier is given in Table 1 of the
        `CMIP6 Data Reference Syntax <https://goo.gl/v1drZl>`_.

        Parameters
        ----------
        root_dir : str, optional
            The root directory of the database i.e. where the cube should start its
            path from e.g. ``/home/users/usertim/cmip6_data``.

        mip_era : str, optional
            The mip_era for which we want to load data e.g. ``CMIP6``.

        activity_id : str, optional
            The activity for which we want to load data e.g. ``DCPP``.

        institution_id : str, optional
            The institution for which we want to load data e.g. ``CNRM-CERFACS``

        source_id : str, optional
            The source_id for which we want to load data e.g. ``CNRM-CM6-1``. This was
            known as model in CMIP5.

        experiment_id : str, optional
            The experiment_id for which we want to load data e.g. ``dcppA-hindcast``.

        member_id : str, optional
            The member_id for which we want to load data e.g. ``s1960-r2i1p1f3``.

        table_id : str, optional
            The table_id for which we want to load data. e.g. ``day``.

        variable_id : str, optional
            The variable_id for which we want to load data e.g. ``pr``.

        grid_label : str, optional
            The grid_label for which we want to load data e.g. ``gn``.

        version : str, optional
            The version for which we want to load data e.g. ``v20160215``.

        time_range : str, optional
            The time range for which we want to load data e.g. ``198001-198412``. If
            ``None``, this information isn't included in the filename which is useful
            for loading metadata files which don't have a relevant time period.

        file_ext : str, optional
            The file extension of the data file we want to load e.g. ``.nc``.

        Returns
        -------
        str
            The full filepath (path and name) of the file to load.
        """
        inargs = locals()
        del inargs["self"]
        # if the step above ever gets more complicated, use the solution here
        # http://kbyanc.blogspot.com/2007/07/python-aggregating-function-arguments.html

        for name, value in inargs.items():
            setattr(self, name, value)

        return join(self._get_data_directory(), self._get_data_filename())

    def get_variable_constraint_from_load_data_from_identifiers_args(
        self, variable_id="tas", **kwargs
    ):
        """Get the iris variable constraint to use when loading data with ``self.load_data_from_identifiers``.

        Parameters
        ----------
        root_dir : str, optional
            The root directory of the database i.e. where the cube should start its
            path from e.g. ``/home/users/usertim/cmip5_25x25``.

        mip_era : str, optional
            The mip_era for which we want to load data.

        activity_id : str, optional
            The activity_id for which we want to load data. For these cubes, will
            almost always be "input4MIPs".

        institution_id : str, optional
            The institution_id for which we want to load data.

        source_id : str, optional
            The source_id for which we want to load data. This was known as model in
            CMIP5.

        experiment_id : str, optional
            The experiment_id for which we want to load data.

        member_id : str, optional
            The member_id for which we want to load data.

        table_id : str, optional
            The table_id for which we want to load data.

        variable_id : str, optional
            The variable_id for which we want to load data.

        grid_label : str, optional
            The grid_label for which we want to load data.

        version : str, optional
            The version for which we want to load data.

        time_range : str, optional
            The time range for which we want to load data. If ``None``, this
            information isn't included in the filename which is useful for loading
            metadata files which don't have a relevant time period.

        file_ext : str, optional
            The file extension of the data file we want to load.

        Returns
        -------
        :obj:`iris.Constraint`
            constraint to use which ensures that only the variable of interest is loaded.
        """
        # thank you Duncan!!
        # https://github.com/SciTools/iris/issues/2107#issuecomment-246644471
        return iris.Constraint(
            cube_func=(lambda c: c.var_name == np.str(variable_id.replace("-", "_")))
        )

    def _get_metadata_load_arguments(self, metadata_variable):
        return {
            "root_dir": self.root_dir,
            "mip_era": self.mip_era,
            "activity_id": self.activity_id,
            "institution_id": self.institution_id,
            "source_id": self.source_id,
            "experiment_id": self.experiment_id,
            "member_id": "r0i0p0",
            "table_id": "fx",
            "variable_id": metadata_variable,
            "grid_label": self.grid_label,
            "version": self.version,
            "time_range": None,
            "file_ext": self.file_ext,
        }

    def _get_data_filename(self):
        bits_to_join = [
            self.variable_id,
            self.table_id,
            self.source_id,
            self.experiment_id,
            self.member_id,
            self.grid_label,
        ]
        if self.time_range is not None:
            bits_to_join.append(self.time_range)

        return "_".join(bits_to_join) + self.file_ext

    def _get_data_directory(self):
        return join(
            self.root_dir,
            self.mip_era,
            self.activity_id,
            self.institution_id,
            self.source_id,
            self.experiment_id,
            self.member_id,
            self.variable_id,
            self.grid_label,
            self.version,
        )



from os.path import join
from unittest.mock import patch, MagicMock
import re
import warnings
import datetime
from dateutil import parser


import pytest
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
import iris
from iris.util import broadcast_to_shape
import cf_units
import cftime
from openscm.scmdataframe import ScmDataFrame


from netcdf_scm.iris_cube_wrappers import (
    SCMCube,
    MarbleCMIP5Cube,
    CMIP6Input4MIPsCube,
    CMIP6OutputCube,
)
from conftest import (
    TEST_TAS_FILE,
    TEST_SFTLF_FILE,
    TEST_AREACELLA_FILE,
    TEST_ACCESS_CMIP5_FILE,
    tdata_required,
    TEST_DATA_MARBLE_CMIP5_DIR,
    TEST_CMIP6_HISTORICAL_CONCS_FILE,
)


class _SCMCubeIntegrationTester(object):
    @tdata_required
    def test_load_data_from_identifiers_and_areacella(self, test_cube):
        tfile = TEST_TAS_FILE
        test_cube.get_filepath_from_load_data_from_identifiers_args = MagicMock(
            return_value=tfile
        )

        test_constraint = iris.Constraint(
            cube_func=(lambda c: c.var_name == np.str("tas"))
        )
        test_cube.get_variable_constraint_from_load_data_from_identifiers_args = MagicMock(
            return_value=test_constraint
        )

        tmdata_scmcube = type(test_cube)()
        tmdata_scmcube.cube = iris.load_cube(TEST_AREACELLA_FILE)
        test_cube.get_metadata_cube = MagicMock(return_value=tmdata_scmcube)

        tkwargs = {
            "variable_name": "fco2antt",
            "modeling_realm": "Amon",
            "model": "CanESM2",
            "experiment": "1pctCO2",
        }

        with pytest.warns(None) as record:
            test_cube.load_data_from_identifiers(**tkwargs)

        assert len(record) == 0

        test_cube.get_filepath_from_load_data_from_identifiers_args.assert_called_with(
            **tkwargs
        )
        test_cube.get_variable_constraint_from_load_data_from_identifiers_args.assert_called_with(
            **tkwargs
        )
        test_cube.get_metadata_cube.assert_called_with(test_cube.areacella_var)

        cell_measures = test_cube.cube.cell_measures()
        assert len(cell_measures) == 1
        assert cell_measures[0].standard_name == "cell_area"

    @tdata_required
    def test_get_scm_timeseries_no_areacealla(self, test_cube):
        var = self.tclass()
        var.cube = iris.load_cube(TEST_TAS_FILE)

        sftlf = self.tclass()
        sftlf.cube = iris.load_cube(TEST_SFTLF_FILE)

        var.get_scm_timeseries(
            sftlf_cube=sftlf, land_mask_threshold=50, areacella_scmcube=None
        )

    def test_get_scm_timeseries_cubes(self, test_cube):
        tsftlf_cube = "mocked 124"
        tland_mask_threshold = "mocked 51"
        tareacella_scmcube = "mocked 4389"

        land_mask = np.array(
            [
                [False, True, True, False],
                [False, True, False, True],
                [False, False, True, False],
            ]
        )
        nh_mask = np.array(
            [
                [False, False, False, False],
                [False, False, False, False],
                [True, True, True, True],
            ]
        )

        mocked_masks = {
            "World": np.full(nh_mask.shape, False),
            "World|Northern Hemisphere|Land": np.logical_or(nh_mask, land_mask),
            "World|Southern Hemisphere|Land": np.logical_or(~nh_mask, land_mask),
            "World|Northern Hemisphere|Ocean": np.logical_or(nh_mask, ~land_mask),
            "World|Southern Hemisphere|Ocean": np.logical_or(~nh_mask, ~land_mask),
        }
        test_cube._get_scm_masks = MagicMock(return_value=mocked_masks)

        mocked_weights = broadcast_to_shape(
            np.array([[1, 2, 3, 4], [1, 4, 8, 9], [0, 4, 1, 9]]),
            test_cube.cube.shape,
            [test_cube.lat_dim_number, test_cube.lon_dim_number],
        )

        test_cube._get_area_weights = MagicMock(return_value=mocked_weights)

        expected = {}
        for label, mask in mocked_masks.items():
            exp_cube = type(test_cube)()

            rcube = test_cube.cube.copy()
            rcube.data.mask = mask
            exp_cube.cube = rcube.collapsed(
                ["latitude", "longitude"], iris.analysis.MEAN, weights=mocked_weights
            )
            expected[label] = exp_cube

        result = test_cube.get_scm_timeseries_cubes(
            tsftlf_cube, tland_mask_threshold, tareacella_scmcube
        )

        for label, cube in expected.items():
            assert cube.cube == result[label].cube

        test_cube._get_scm_masks.assert_called_with(
            sftlf_cube=tsftlf_cube, land_mask_threshold=tland_mask_threshold
        )
        test_cube._get_area_weights.assert_called_with(
            areacella_scmcube=tareacella_scmcube
        )

    @pytest.mark.parametrize("out_calendar", [None, "gregorian", "365_day"])
    def test_convert_scm_timeseries_cubes_to_openscmdata(self, test_cube, out_calendar):
        expected_calendar = (
            test_cube.cube.coords("time")[0].units.calendar
            if out_calendar is None
            else out_calendar
        )

        global_cube = type(test_cube)()
        global_cube.cube = test_cube.cube.copy()
        global_cube.cube.data = 2 * global_cube.cube.data

        # can safely ignore warnings here
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", ".*without weighting*")
            global_cube.cube = global_cube.cube.collapsed(
                ["longitude", "latitude"], iris.analysis.MEAN
            )

        sh_ocean_cube = type(test_cube)()
        sh_ocean_cube.cube = test_cube.cube.copy()
        sh_ocean_cube.cube.data = 0.5 * sh_ocean_cube.cube.data
        with warnings.catch_warnings(record=True):
            warnings.filterwarnings("ignore", ".*without weighting*")
            sh_ocean_cube.cube = sh_ocean_cube.cube.collapsed(
                ["longitude", "latitude"], iris.analysis.MEAN
            )

        test_timeseries_cubes = {
            "World": global_cube,
            "World|Southern Hemisphere|Ocean": sh_ocean_cube,
        }
        with warnings.catch_warnings(record=True):
            warnings.filterwarnings("ignore", ".*appropriate model scenario*")
            result = test_cube._convert_scm_timeseries_cubes_to_openscmdata(
                test_timeseries_cubes, out_calendar=out_calendar
            )

        time = sh_ocean_cube.cube.dim_coords[0]
        datetimes = cf_units.num2date(time.points, time.units.name, expected_calendar)
        if not isinstance(datetimes[0], datetime.datetime):
            datetimes = np.array([parser.parse(x.strftime()) for x in datetimes])
        time_index = pd.Index(datetimes, dtype="object", name="time")

        expected_df = pd.DataFrame(
            {
                "World": global_cube.cube.data,
                "World|Southern Hemisphere|Ocean": sh_ocean_cube.cube.data,
            },
            index=time_index,
        )

        expected_df.columns = pd.MultiIndex.from_product(
            [
                [test_cube.cube.standard_name],
                [test_cube.cube.units.name],
                expected_df.columns.tolist(),
                ["unspecified"],
                ["unspecified"],
                ["unspecified"],
            ],
            names=["variable", "unit", "region", "climate_model", "scenario", "model"],
        )
        expected_df = (
            expected_df.unstack().reset_index().rename({0: "value"}, axis="columns")
        )

        expected = ScmDataFrame(expected_df)
        expected.metadata = {"calendar": expected_calendar}

        assert result.metadata == expected.metadata
        assert_frame_equal(result.timeseries(), expected.timeseries())

    @patch("netcdf_scm.iris_cube_wrappers.os.listdir")
    def test_check_data_names_in_same_directory(self, mock_listdir, test_cube):
        tdir = "mocked"

        mock_listdir.return_value = [
            "tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc",
            "tas_Amon_HadCM3_rcp45_r1i1p1_203101-203512.nc",
            "tas_Amon_HadCM3_rcp45_r1i1p1_203601-203812.nc",
        ]

        test_cube._check_data_names_in_same_directory(tdir)
        mock_listdir.assert_called_with(tdir)

    @patch("netcdf_scm.iris_cube_wrappers.os.listdir")
    @pytest.mark.parametrize(
        "bad_file_list",
        [
            [
                "tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc",
                "pr_Amon_HadCM3_rcp45_r1i1p1_203101-203512.nc",
            ],
            [
                "tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc",
                "tas_fx_HadCM3_rcp45_r1i1p1_203101-203512.nc",
            ],
            [
                "tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc",
                "tas_Amon_CSIRO_rcp45_r1i1p1_203101-203512.nc",
            ],
            [
                "tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc",
                "tas_Amon_HadCM3_rcp85_r1i1p1_203101-203512.nc",
            ],
            [
                "tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc",
                "tas_Amon_HadCM3_rcp45_r2i1p1_203101-203512.nc",
            ],
            [
                "tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc",
                "tas_Amon_HadCM3_rcp45_r1i1p1_203201-203512.nc",
            ],
            [
                "tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc",
                "tas_Amon_HadCM3_rcp45_r1i1p1_203012-203512.nc",
            ],
            [
                "tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc",
                "pr_Amon_HadCM3_rcp45_r1i1p1_203101-203412.nc",
                "tas_Amon_HadCM3_rcp45_r1i1p1_203601-203812.nc",
            ],
            [
                "tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc",
                "tas_Amon_HadCM3_rcp45_r1i1p1_203101-203512.nc",
                "tas_Amon_HadCM3_rcp45_r1i1p1_203701-203812.nc",
            ],
            [
                "tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc",
                "tas_Amon_HadCM3_rcp45_r1i1p1_203101-203512.nc",
                "tas_Amon_HadCM3_rcp45_r1i1p1_203602-203812.nc",
            ],
            [
                "tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc",
                "tas_Amon_HadCM3_rcp45_r1i1p1_203101-203512.nc",
                "tas_Amon_HadCM3_rcp45_r1i1p2_203601-203812.nc",
            ],
        ],
    )
    def test_check_data_names_in_same_directory_errors(
        self, mock_listdir, bad_file_list, test_cube
    ):
        tdir = "mocked"

        mock_listdir.return_value = bad_file_list
        error_msg = re.escape(
            (
                "Cannot join files in:\n"
                "{}\n"
                "Files found:\n"
                "- {}".format(tdir, "\n- ".join(sorted(bad_file_list)))
            )
        )
        with pytest.raises(AssertionError, match=error_msg):
            test_cube._check_data_names_in_same_directory(tdir)


class TestSCMCubeIntegration(_SCMCubeIntegrationTester):
    tclass = SCMCube

    def test_load_and_concatenate_files_in_directory_same_time(self, test_cube):
        tdir = join(
            TEST_DATA_MARBLE_CMIP5_DIR,
            "cmip5",
            "rcp45",
            "Amon",
            "tas",
            "HadCM3",
            "r1i1p1",
        )

        # can ignore warnings safely here as tested elsewhere
        with warnings.catch_warnings(record=True):
            test_cube._load_and_concatenate_files_in_directory(tdir)

        obs_time = test_cube.cube.dim_coords[0]
        obs_time = cf_units.num2date(
            obs_time.points, obs_time.units.name, obs_time.units.calendar
        )
        assert obs_time[0] == cftime.Datetime360Day(2006, 1, 16, 0, 0, 0, 0, -1, 16)
        assert obs_time[-1] == cftime.Datetime360Day(2035, 12, 16, 0, 0, 0, 0, -1, 346)

        removed_attributes = ["creation_date", "tracking_id", "history"]
        for removed_attribute in removed_attributes:
            with pytest.raises(KeyError):
                test_cube.cube.attributes[removed_attribute]

    def test_load_and_concatenate_files_in_directory_different_time(self, test_cube):
        tdir = join(
            TEST_DATA_MARBLE_CMIP5_DIR,
            "cmip5",
            "rcp85",
            "Amon",
            "tas",
            "NorESM1-ME",
            "r1i1p1",
        )

        with warnings.catch_warnings(record=True):
            test_cube._load_and_concatenate_files_in_directory(tdir)

        obs_time = test_cube.cube.dim_coords[0]
        obs_time = cf_units.num2date(
            obs_time.points, obs_time.units.name, obs_time.units.calendar
        )
        assert obs_time[0] == cftime.DatetimeNoLeap(2006, 1, 16, 12, 0, 0, 0, 0, 16)
        assert obs_time[-1] == cftime.DatetimeNoLeap(2100, 12, 16, 12, 0, 0, 0, 1, 350)

        removed_attributes = ["creation_date", "tracking_id", "history"]
        for removed_attribute in removed_attributes:
            with pytest.raises(KeyError):
                test_cube.cube.attributes[removed_attribute]

    def test_load_gregorian_calendar_with_pre_zero_years(self, test_cube):
        expected_warn = (
            "Your calendar is gregorian yet has units of 'days since 0-1-1'. We "
            "rectify this by removing all data before year 1 and changing the units "
            "to 'days since 1-1-1'. If you want other behaviour, you will need to use "
            "another package."
        )
        with warnings.catch_warnings(record=True) as adjust_warnings:
            test_cube.load_data_from_path(TEST_CMIP6_HISTORICAL_CONCS_FILE)

        assert len(adjust_warnings) == 1
        assert str(adjust_warnings[0].message) == expected_warn

        obs_time = test_cube.cube.dim_coords[0]
        assert obs_time.units.name == "day since 1-01-01 00:00:00.000000 UTC"
        assert obs_time.units.calendar == "gregorian"

        obs_time_points = cf_units.num2date(
            obs_time.points, obs_time.units.name, obs_time.units.calendar
        )
        assert obs_time_points[0] == datetime.datetime(1, 7, 3, 12, 0)
        assert obs_time_points[-1] == datetime.datetime(2014, 7, 3, 12, 0)

        assert test_cube.cube.attributes["institution_id"] == "UoM"
        assert test_cube.cube.attributes["Conventions"] == "CF-1.6"
        assert test_cube.cube.attributes["table_id"] == "input4MIPs"
        assert test_cube.cube.cell_methods[0].method == "mean"
        assert str(test_cube.cube.units) == "1.e-12"
        assert test_cube.cube.var_name == "mole_fraction_of_so2f2_in_air"
        assert test_cube.cube.name() == "mole"
        assert test_cube.cube.long_name == "mole"
        assert isinstance(test_cube.cube.metadata, iris.cube.CubeMetadata)

    def test_access_cmip5_read_issue_30(self, test_cube):
        test_cube.load_data_from_path(TEST_ACCESS_CMIP5_FILE)

        obs_time = test_cube.cube.dim_coords[0]
        assert obs_time.units.name == "day since 1-01-01 00:00:00.000000 UTC"
        assert obs_time.units.calendar == "proleptic_gregorian"

        obs_time_points = cf_units.num2date(
            obs_time.points, obs_time.units.name, obs_time.units.calendar
        )
        assert obs_time_points[0] == datetime.datetime(2006, 1, 16, 12, 0)
        assert obs_time_points[-1] == datetime.datetime(2049, 12, 16, 12, 0)


class TestMarbleCMIP5Cube(_SCMCubeIntegrationTester):
    tclass = MarbleCMIP5Cube

    def test_load_and_concatenate_files_in_directory_same_time(self, test_cube):
        tdir = join(
            TEST_DATA_MARBLE_CMIP5_DIR,
            "cmip5",
            "rcp45",
            "Amon",
            "tas",
            "HadCM3",
            "r1i1p1",
        )

        # can ignore warnings safely here as tested elsewhere
        with warnings.catch_warnings(record=True):
            test_cube._load_and_concatenate_files_in_directory(tdir)

        obs_time = test_cube.cube.dim_coords[0]
        obs_time = cf_units.num2date(
            obs_time.points, obs_time.units.name, obs_time.units.calendar
        )
        assert obs_time[0] == cftime.Datetime360Day(2006, 1, 16, 0, 0, 0, 0, -1, 16)
        assert obs_time[-1] == cftime.Datetime360Day(2035, 12, 16, 0, 0, 0, 0, -1, 346)

        assert test_cube.time_period == "200601-203512"

        removed_attributes = ["creation_date", "tracking_id", "history"]
        for removed_attribute in removed_attributes:
            with pytest.raises(KeyError):
                test_cube.cube.attributes[removed_attribute]

    def test_load_and_concatenate_files_in_directory_different_time(self, test_cube):
        tdir = join(
            TEST_DATA_MARBLE_CMIP5_DIR,
            "cmip5",
            "rcp85",
            "Amon",
            "tas",
            "NorESM1-ME",
            "r1i1p1",
        )

        test_cube._load_and_concatenate_files_in_directory(tdir)

        obs_time = test_cube.cube.dim_coords[0]
        obs_time = cf_units.num2date(
            obs_time.points, obs_time.units.name, obs_time.units.calendar
        )
        assert obs_time[0] == cftime.DatetimeNoLeap(2006, 1, 16, 12, 0, 0, 0, 0, 16)
        assert obs_time[-1] == cftime.DatetimeNoLeap(2100, 12, 16, 12, 0, 0, 0, 1, 350)

        assert test_cube.time_period == "200601-210012"

        removed_attributes = ["creation_date", "tracking_id", "history"]
        for removed_attribute in removed_attributes:
            with pytest.raises(KeyError):
                test_cube.cube.attributes[removed_attribute]

    def test_get_load_data_from_identifiers_args_from_filepath(self, test_cube):
        tpath = "tests/test_data/marble_cmip5/cmip5/1pctCO2/Amon/fco2antt/CanESM2/r1i1p1/fco2antt_Amon_CanESM2_1pctCO2_r1i1p1_185001-198912.nc"
        expected = {
            "root_dir": "tests/test_data/marble_cmip5",
            "activity": "cmip5",
            "experiment": "1pctCO2",
            "modeling_realm": "Amon",
            "variable_name": "fco2antt",
            "model": "CanESM2",
            "ensemble_member": "r1i1p1",
            "time_period": "185001-198912",
            "file_ext": ".nc",
        }
        result = test_cube.get_load_data_from_identifiers_args_from_filepath(tpath)

        assert result == expected

    def test_get_load_data_from_identifiers_args_from_filepath_no_time(self, test_cube):
        tpath = "tests/test_data/marble_cmip5/cmip5/1pctCO2/fx/sftlf/CanESM2/r0i0p0/sftlf_fx_CanESM2_1pctCO2_r0i0p0.nc"
        expected = {
            "root_dir": "tests/test_data/marble_cmip5",
            "activity": "cmip5",
            "experiment": "1pctCO2",
            "modeling_realm": "fx",
            "variable_name": "sftlf",
            "model": "CanESM2",
            "ensemble_member": "r0i0p0",
            "time_period": None,
            "file_ext": ".nc",
        }
        result = test_cube.get_load_data_from_identifiers_args_from_filepath(tpath)

        assert result == expected

    def test_get_load_data_from_identifiers_args_from_filepath_no_root_dir(
        self, test_cube
    ):
        tpath = (
            "cmip5/1pctCO2/fx/sftlf/CanESM2/r0i0p0/sftlf_fx_CanESM2_1pctCO2_r0i0p0.nc"
        )
        expected = {
            "root_dir": ".",
            "activity": "cmip5",
            "experiment": "1pctCO2",
            "modeling_realm": "fx",
            "variable_name": "sftlf",
            "model": "CanESM2",
            "ensemble_member": "r0i0p0",
            "time_period": None,
            "file_ext": ".nc",
        }
        result = test_cube.get_load_data_from_identifiers_args_from_filepath(tpath)

        assert result == expected

    @pytest.mark.parametrize(
        "tpath",
        ["cmip5/1pctCO2/fx/sftlf/CanESM2/r0i0p0/sftlf_fx_HadGem3_1pctCO2_r0i0p0.nc"],
    )
    def test_get_load_data_from_identifiers_args_from_filepath_errors(
        self, test_cube, tpath
    ):
        error_msg = (
            re.escape("Path and filename do not agree:")
            + "\n"
            + re.escape("    - path model: CanESM2")
            + "\n"
            + re.escape("    - filename model: HadGem3")
            + "\n"
        )
        with pytest.raises(ValueError, match=error_msg):
            test_cube.get_load_data_from_identifiers_args_from_filepath(tpath)

    def test_access_cmip5_read_issue_30(self, test_cube):
        test_cube.load_data_from_path(TEST_ACCESS_CMIP5_FILE)

        obs_time = test_cube.cube.dim_coords[0]
        assert obs_time.units.name == "day since 1-01-01 00:00:00.000000 UTC"
        assert obs_time.units.calendar == "proleptic_gregorian"

        obs_time_points = cf_units.num2date(
            obs_time.points, obs_time.units.name, obs_time.units.calendar
        )
        assert obs_time_points[0] == datetime.datetime(2006, 1, 16, 12, 0)
        assert obs_time_points[-1] == datetime.datetime(2049, 12, 16, 12, 0)

        assert test_cube.model == "ACCESS1-0"


class TestCMIP6Input4MIPsCube(_SCMCubeIntegrationTester):
    tclass = CMIP6Input4MIPsCube

    def test_load_gregorian_calendar_with_pre_zero_years(self, test_cube):
        expected_warn = (
            "Your calendar is gregorian yet has units of 'days since 0-1-1'. We "
            "rectify this by removing all data before year 1 and changing the units "
            "to 'days since 1-1-1'. If you want other behaviour, you will need to use "
            "another package."
        )
        with warnings.catch_warnings(record=True) as adjust_warnings:
            test_cube.load_data_from_path(TEST_CMIP6_HISTORICAL_CONCS_FILE)

        assert len(adjust_warnings) == 1
        assert str(adjust_warnings[0].message) == expected_warn

        obs_time = test_cube.cube.dim_coords[0]
        assert obs_time.units.name == "day since 1-01-01 00:00:00.000000 UTC"
        assert obs_time.units.calendar == "gregorian"

        obs_time_points = cf_units.num2date(
            obs_time.points, obs_time.units.name, obs_time.units.calendar
        )

        assert obs_time_points[0] == datetime.datetime(1, 7, 3, 12, 0)
        assert obs_time_points[-1] == datetime.datetime(2014, 7, 3, 12, 0)

        assert test_cube.cube.attributes["institution_id"] == "UoM"
        assert test_cube.cube.attributes["Conventions"] == "CF-1.6"
        assert test_cube.cube.attributes["table_id"] == "input4MIPs"
        assert test_cube.cube.cell_methods[0].method == "mean"
        assert str(test_cube.cube.units) == "1.e-12"
        assert test_cube.cube.var_name == "mole_fraction_of_so2f2_in_air"
        assert test_cube.cube.name() == "mole"
        assert test_cube.cube.long_name == "mole"
        assert isinstance(test_cube.cube.metadata, iris.cube.CubeMetadata)

    def test_get_load_data_from_identifiers_args_from_filepath(self, test_cube):
        tpath = "tests/test_data/cmip6-input4mips/input4MIPs/CMIP6/CMIP/PCMDI/PCMDI-AMIP-1-1-4/ocean/mon/tos/gn/v20180427/tos_input4MIPs_SSTsAndSeaIce_CMIP_PCMDI-AMIP-1-1-4_gn_187001-201712.nc"
        expected = {
            "root_dir": "tests/test_data/cmip6-input4mips",
            "activity_id": "input4MIPs",
            "mip_era": "CMIP6",
            "target_mip": "CMIP",
            "institution_id": "PCMDI",
            "source_id": "PCMDI-AMIP-1-1-4",
            "realm": "ocean",
            "frequency": "mon",
            "variable_id": "tos",
            "grid_label": "gn",
            "version": "v20180427",
            "dataset_category": "SSTsAndSeaIce",
            "time_range": "187001-201712",
            "file_ext": ".nc",
        }
        result = test_cube.get_load_data_from_identifiers_args_from_filepath(tpath)

        assert result == expected

    def test_get_load_data_from_identifiers_args_from_filepath_no_time(self, test_cube):
        tpath = "tests/test_data/cmip6-input4mips/input4MIPs/CMIP6/CMIP/PCMDI/PCMDI-AMIP-1-1-4/land/fx/sftlf/gn/v20180427/sftlf_input4MIPs_landState_CMIP_PCMDI-AMIP-1-1-4_gn.nc"
        expected = {
            "root_dir": "tests/test_data/cmip6-input4mips",
            "activity_id": "input4MIPs",
            "mip_era": "CMIP6",
            "target_mip": "CMIP",
            "institution_id": "PCMDI",
            "source_id": "PCMDI-AMIP-1-1-4",
            "realm": "land",
            "frequency": "fx",
            "variable_id": "sftlf",
            "grid_label": "gn",
            "version": "v20180427",
            "dataset_category": "landState",
            "time_range": None,
            "file_ext": ".nc",
        }
        result = test_cube.get_load_data_from_identifiers_args_from_filepath(tpath)

        assert result == expected

    def test_get_load_data_from_identifiers_args_from_filepath_no_root_dir(
        self, test_cube
    ):
        tpath = "input4MIPs/CMIP6/CMIP/PCMDI/PCMDI-AMIP-1-1-4/ocean/mon/tos/gn/v20180427/tos_input4MIPs_SSTsAndSeaIce_CMIP_PCMDI-AMIP-1-1-4_gn_187001-201712.nc"
        expected = {
            "root_dir": ".",
            "activity_id": "input4MIPs",
            "mip_era": "CMIP6",
            "target_mip": "CMIP",
            "institution_id": "PCMDI",
            "source_id": "PCMDI-AMIP-1-1-4",
            "realm": "ocean",
            "frequency": "mon",
            "variable_id": "tos",
            "grid_label": "gn",
            "version": "v20180427",
            "dataset_category": "SSTsAndSeaIce",
            "time_range": "187001-201712",
            "file_ext": ".nc",
        }
        result = test_cube.get_load_data_from_identifiers_args_from_filepath(tpath)

        assert result == expected

    @pytest.mark.parametrize(
        "tpath",
        [
            "input4MIPs/CMIP6/CMIP/PCMDI/PCMDI-AMIP-1-1-4/ocean/mon/tas/gn/v20180427/tos_input4MIPs_SSTsAndSeaIce_CMIP_PCMDI-AMIP-1-1-4_gn_187001-201712.nc"
        ],
    )
    def test_get_load_data_from_identifiers_args_from_filepath_errors(
        self, test_cube, tpath
    ):
        error_msg = (
            re.escape("Path and filename do not agree:")
            + "\n"
            + re.escape("    - path variable_id: tas")
            + "\n"
            + re.escape("    - filename variable_id: tos")
            + "\n"
        )
        with pytest.raises(ValueError, match=error_msg):
            test_cube.get_load_data_from_identifiers_args_from_filepath(tpath)


class TestCMIP6OutputCube(_SCMCubeIntegrationTester):
    tclass = CMIP6OutputCube

    def test_get_load_data_from_identifiers_args_from_filepath(self, test_cube):
        tpath = "tests/test_data/cmip6-output/CMIP6/DCPP/CNRM-CERFACS/CNRM-CM6-1/dcppA-hindcast/s1960-r2i1p1f3/day/pr/gn/v20160215/pr_day_CNRM-CM6-1_dcppA-hindcast_s1960-r2i1p1f3_gn_198001-198412.nc"
        expected = {
            "root_dir": "tests/test_data/cmip6-output",
            "mip_era": "CMIP6",
            "activity_id": "DCPP",
            "institution_id": "CNRM-CERFACS",
            "source_id": "CNRM-CM6-1",
            "experiment_id": "dcppA-hindcast",
            "member_id": "s1960-r2i1p1f3",
            "table_id": "day",
            "variable_id": "pr",
            "grid_label": "gn",
            "version": "v20160215",
            "time_range": "198001-198412",
            "file_ext": ".nc",
        }
        result = test_cube.get_load_data_from_identifiers_args_from_filepath(tpath)

        assert result == expected

    def test_get_load_data_from_identifiers_args_from_filepath_no_time(self, test_cube):
        tpath = "tests/test_data/cmip6-output/CMIP6/DCPP/CNRM-CERFACS/CNRM-CM6-1/dcppA-hindcast/r0i0p0/fx/sftlf/gn/v20160215/sftlf_fx_CNRM-CM6-1_dcppA-hindcast_r0i0p0_gn.nc"
        expected = {
            "root_dir": "tests/test_data/cmip6-output",
            "mip_era": "CMIP6",
            "activity_id": "DCPP",
            "institution_id": "CNRM-CERFACS",
            "source_id": "CNRM-CM6-1",
            "experiment_id": "dcppA-hindcast",
            "member_id": "r0i0p0",
            "table_id": "fx",
            "variable_id": "sftlf",
            "grid_label": "gn",
            "version": "v20160215",
            "time_range": None,
            "file_ext": ".nc",
        }
        result = test_cube.get_load_data_from_identifiers_args_from_filepath(tpath)

        assert result == expected

    def test_get_load_data_from_identifiers_args_from_filepath_no_root_dir(
        self, test_cube
    ):
        tpath = "CMIP6/DCPP/CNRM-CERFACS/CNRM-CM6-1/dcppA-hindcast/s1960-r2i1p1f3/day/pr/gn/v20160215/pr_day_CNRM-CM6-1_dcppA-hindcast_s1960-r2i1p1f3_gn_198001-198412.nc"
        expected = {
            "root_dir": ".",
            "mip_era": "CMIP6",
            "activity_id": "DCPP",
            "institution_id": "CNRM-CERFACS",
            "source_id": "CNRM-CM6-1",
            "experiment_id": "dcppA-hindcast",
            "member_id": "s1960-r2i1p1f3",
            "table_id": "day",
            "variable_id": "pr",
            "grid_label": "gn",
            "version": "v20160215",
            "time_range": "198001-198412",
            "file_ext": ".nc",
        }
        result = test_cube.get_load_data_from_identifiers_args_from_filepath(tpath)

        assert result == expected

    @pytest.mark.parametrize(
        "tpath",
        [
            "CMIP6/DCPP/CNRM-CERFACS/CNRM-CM6-1/dcppA-hindcast/s1960-r2i1p1f3/day/pr/gn/v20160215/pr_day_CNRM-CM6-1_dcppA_s1960-r2i1p1f3_gn_198001-198412.nc"
        ],
    )
    def test_get_load_data_from_identifiers_args_from_filepath_errors(
        self, test_cube, tpath
    ):
        error_msg = (
            re.escape("Path and filename do not agree:")
            + "\n"
            + re.escape("    - path experiment_id: dcppA-hindcast")
            + "\n"
            + re.escape("    - filename experiment_id: dcppA")
            + "\n"
        )
        with pytest.raises(ValueError, match=error_msg):
            test_cube.get_load_data_from_identifiers_args_from_filepath(tpath)
